#include "stdafx.h"
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdarg.h>
#include "Ygl.h"
#include "syntax.h"
#include "box.h"
#include "grid.h"
#include "vector.h"
#include "field.h"
#include "table.h"
#include "peak.h"
#include "interpol.h"
#include "simulation.h"
#include "fplot.h"
#include "gate.h"
#include "loop.h"

double CHARGE_LOSS;

int     VERBOSE = 2;
int     GEOMETRY = CARTESIAN3D;

FILE *stdlog = stderr;

//********************************************************************************************

#define  EXTRA_PARAM_STRING  "; pA=5.182134 ; pi=4 atan(1) ; "

#define RETURN_NORMAL 0     // return types used by ADI step methods
#define RETURN_END    1

//********************************************************************************************

void getRun( TokenString &TS, int i, bool *adaptive=0, char *PDEmethod=0, double *time=0, double *dt=0, 
             double *dtMax=0, double *accuracy=0, double *accuracy0=0, double *maxError=0, double *dtStretch=0, 
             int *n0=0, double *nStretch=0, int *nMax=0, double *ODEaccuracy=0);

const char *getMethod(char *smethod, CaMethod &, BufMethod &);
double get_sim_time(TokenString &params);

void setTortuosity(TokenString &TS, long p, double *Diff);

int Buf1Dstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);
int Ca1Dstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);

int Buf2Dstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);
int Ca2Dstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);

int Buf2DFIstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);
int Ca2DFIstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);

int BufEu2Dstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);
int CaEu2Dstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);

int BufEulerStep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);
int CaEulerStep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);

int CaADIstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);
int BufADIstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);

int CaLinADIstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);
int BufLinADIstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);

int CaFIADIstep(FieldObj &Ca, VectorObj &Canext, BufferArray &, double dt, double);
int BufFIADIstep(BufferObj &Buf, VectorObj &Buf_next, FieldObj &Ca, double dt, double);

//**************************************************************************************************

void header() {
     fprintf(stdlog,"\n*******************************************************");
     fprintf(stdlog,"\n*                                                     *");
     fprintf(stdlog,"\n*  Calcium diffusion/buffering solver: version 4.5.1  *");
     fprintf(stdlog,"\n*   Victor Matveev, MRB/NIDDK/NIH, February 6, 2003   *");
     fprintf(stdlog,"\n*                                                     *");
     fprintf(stdlog,"\n*******************************************************\n");
}

 //**************************************************************************************************
 //                                           M A I N
 //**************************************************************************************************


 int main(int argc, char **argv) {

 char script[512];

 try {

   if (argc < 2) {
     header();
     fprintf(stderr, "\n\nEnter the CalC cript file name: ");
     scanf("%s", script);
   }
   else strcpy(script, argv[1]);

   TokenString *TS;

   TS = new TokenString(script, EXTRA_PARAM_STRING, "", argc, argv);
   TS->get_int_param("verbose", &VERBOSE);
   if (VERBOSE) header();

   VectorObj result( getTrackVarNum(*TS) );
   LoopObj   vary( *TS, &result );
   delete TS;

   // ++++++++++++++++++++++++++++ MAIN LOOP ++++++++++++++++++++++++++++

   for (int step = 0; step < vary.steps; step++) {
     vary.step();
     TS = new TokenString(script, EXTRA_PARAM_STRING, vary.loopVarString, argc, argv); 
     if (TS->token_count("exit")) break;
     TS->get_int_param("verbose", &VERBOSE);
     SimulationObj *Simulation;
     double *trackPtr;

     if ( !TS->token_count("Run") ) {                    //####### CALCULATOR MODE #######
       if (VERBOSE) fprintf(stdlog, "***** No simulation runs specified; running in calculator mode *****\n");
       if ( vary() ) for (int i = 0; i < result.size; i++) 
	  result[i] = TS->get_double( getTrackVar(*TS, i) );
       TS->printResults(0);
     }        
     else {
       if ( equal(TS->get_string("mode"), "ODE") )  {// beware of undestructed string
         if (VERBOSE) fprintf(stdlog, "***** Running in ODE-only mode *****\n");
	 Simulation = new ODESimulationObj(*TS);    //#######  ODE solver mode: ######
       } 
       else  Simulation = new SimulationObj(*TS);       //#######  FULL PDE MODE  #######
       Simulation->Run();                               //##### RUN THE DIFFERENCE SCHEME #####
       if ( vary() ) for (int i = 0; i < result.size; i++) 
         if ( (trackPtr = Simulation->ResolveID( vary.trackIDs[i] )) ) result[i] = *trackPtr;
         else  TS->errorMessage( TS->token_index(vary.trackIDs[i]), 0, "Cannot track an undefined variable" );
       TS->printResults(Simulation);
       delete Simulation;
       }
     vary.draw(); 
     delete TS;
   }

 }
 catch (char *str) {  if ( !equal(str,"") ) {
                        fprintf(stdlog, "\n\n*** Error: %s\n", str);
                        delete [] str;
                      }
                   }
 catch (int i) { return i; }
 return 0;
 }


 //**********************************************************************************************
 //@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
 //**********************************************************************************************

//**************************************************************************
//                C A L C I U M   D - G   S T E P
//**************************************************************************
//
//  (1 - Ax/2) Ca*  = (1 + Ax/2 + Ay + Az) Ca + dt H(Ca,B)
//  (1 - Ay/2) Ca** = -Ay/2 Ca + Ca*
//  (1 - Az/2) Ca^  = -Az/2 Ca + Ca** + dt / 2 * { H(Ca^,B) - H(Ca,B) }
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }
//
//  Ca = Ca(t_n), Ca^ = Ca(t_{n+1}), B = B(t_{n+1/2})
//**************************************************************************

int CaADIstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int       rvalue = RETURN_NORMAL;
  double    t0 = Ca.Time;
  VectorObj Lin(FieldObj::Size);

  if ( (T > 0.0) && (t0 + dt > T))  {  rvalue = RETURN_END; dt = T - t0; }

  double nu = Ca.getD() * dt;
  Next = Ca;
  Lin = 0.0;
  double sum = 0.0;  
  for (int b = 0; b < Buf.buf_num; b++)
    {
    Lin  -= (dt * Buf.array[b]->kplus) * *Buf.array[b];
    Next -= ( dt * Buf.array[b]->kminus) * *Buf.array[b];
    sum += dt * (Buf.array[b])->kminus * (Buf.array[b])->total;
    }

  Next += sum + Lin % Ca;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);

  Ca.xRun(-0.5 * nu, 0.5 * nu,    nu,     nu,  0,  Next.elem);
  Ca.yRun(-0.5 * nu,     0.,  -0.5 * nu,  0.,  0,  Next.elem);

  Lin *= 0.5;
  Next -= (Lin % Ca);

  Ca.zRun(-0.5 * nu, 0., 0., -0.5 * nu, Lin.elem, Next.elem);

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; /*Ca.cleanBoundaries();*/ }
    else Ca.Time = t0;

  return rvalue;
}

//**************************************************************************
//                  B U F F E R   D - G   S T E P
//**************************************************************************

int BufADIstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{ 
int    rvalue = RETURN_NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) 
    { dt = T - Buf.Time; /*printf("Buf: dt reduced to %g\n", dt);*/ rvalue = RETURN_END; }
  double nu = Buf.getD() * dt;
  
  VectorObj Lin( -dt * (Buf.kplus * Ca + Buf.kminus ) );
  Next = Buf % ( 1.0 + Lin ) + dt * Buf.kminus * Buf.total;

  Buf.xRun(-0.5 * nu, 0.5 * nu, nu, nu, 0, Next.elem);
  Buf.yRun(-0.5 * nu, 0., -0.5 * nu, 0., 0, Next.elem);

  Lin *= 0.5;
  Next -= (Lin % Buf);
  Buf.zRun(-0.5 * nu, 0., 0., -0.5 * nu,  Lin.elem, Next.elem);

  if (T > 0.0) { Buf.Time += dt; Buf = Next;  /*Buf.cleanBoundaries();*/ }
  
  return rvalue;
}

//**************************************************************************
//                C A L C I U M   2 D   D - G   S T E P
//**************************************************************************
//
//  (1 - Ax/2) Ca*  = (1 + Ay/2) Ca  + dt/2 H(Ca,B)
//  (1 - Ay/2) Ca^  = (1 + Ax/2) Ca* + dt/2 H(Ca^,B)
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }
//
//  Ca = Ca(t_n), Ca^ = Ca(t_{n+1}), B = B(t_{n+1/2})
//**************************************************************************

int Ca2Dstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int       b, rvalue = RETURN_NORMAL;
  double    t0 = Ca.Time;
  VectorObj Lin(FieldObj::Size);

  if ( (T > 0.0) && (t0 + dt > T))  { rvalue = RETURN_END; dt = T - t0; }

  double nu = Ca.getD() * dt;

  Next = Ca;
  Lin = 0.0; 
  double sum = 0.0;  
  for (b = 0; b < Buf.buf_num; b++)
    {
    Lin  -= (0.5 * dt * Buf.array[b]->kplus) * *Buf.array[b];
    Next -= *Buf.array[b] * (0.5 * dt * Buf.array[b]->kminus);
    sum += 0.5 * dt * (Buf.array[b])->kminus * (Buf.array[b])->total;
    }

  Next += sum + Lin % Ca;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, 0.5 * dt);

  //  void (FieldObj::*routine)(double, double, double, double *, double *);
  //routine = &FieldObj::RunConR;

  //Ca.routine(    -0.5 * nu,     0.0,    0.5 * nu,  0,  Next.elem);

  switch (GEOMETRY) {
  case CONICAL:  Ca.RunConR(    -0.5 * nu,     0.0,    0.5 * nu,  0,  Next.elem); break;
  case CYLINDRICAL:  Ca.RunCylR(    -0.5 * nu,     0.0,    0.5 * nu,  0,  Next.elem); break;
  }

  for (b = 0; b < Buf.buf_num; b++)    
    Next -= *Buf.array[b] * (0.5 * dt * Buf.array[b]->kminus);
  Next += sum;
  Ca.add_sources(Next, 0.5 * dt);

  switch (GEOMETRY) {
  case CONICAL:  Ca.RunConTheta(-0.5 * nu, 0.5 * nu,     0.0,     Lin.elem,  Next.elem); break;
  case CYLINDRICAL:  Ca.RunCylZ(-0.5 * nu, 0.5 * nu,     0.0,     Lin.elem,  Next.elem); break;
  }

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; /*Ca.cleanBoundaries();*/ }
    else Ca.Time = t0;

  return rvalue;
}

//**************************************************************************
//               B U F F E R   2 D   D - G   S T E P
//**************************************************************************
//
//  (1 - Ax/2) B*  = (1 + Ay/2) B  + dt/2 H(Ca,B)
//  (1 - Ay/2) B^  = (1 + Ax/2) B* + dt/2 H(Ca,B^)
//
//  H(Ca,B) = - kplus * Ca * B + kminus * (B_total - B)
//
//  B = B(t_n), B^ = B(t_{n+1}), Ca = Ca(t_{n+1/2})
//**************************************************************************

int Buf2Dstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{ 
int    rvalue = RETURN_NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) 
    { dt = T - Buf.Time; /*printf("Buf: dt reduced to %g\n", dt);*/ rvalue = RETURN_END; }
  double nu = Buf.getD() * dt;
  
  VectorObj Lin( (-0.5 * dt) * (Buf.kplus * Ca + Buf.kminus) );

  Next = Buf + 0.5 * dt * Buf.kminus * Buf.total + Lin % Buf;

  switch (GEOMETRY) {
  case CONICAL:  Buf.RunConR(    -0.5 * nu,     0.0,    0.5 * nu,  0,  Next.elem); break;
  case CYLINDRICAL:  Buf.RunCylR(    -0.5 * nu,     0.0,    0.5 * nu,  0,  Next.elem); break;  
  }

  Next += 0.5 * dt * Buf.kminus * Buf.total;
  
  switch (GEOMETRY) {
  case CONICAL: Buf.RunConTheta(-0.5 * nu, 0.5 * nu,     0.0,     Lin.elem,  Next.elem); break;
  case CYLINDRICAL: Buf.RunCylZ(-0.5 * nu, 0.5 * nu,     0.0,     Lin.elem,  Next.elem); break;
  }

  if (T > 0.0) { Buf.Time += dt; Buf = Next;  /*Buf.cleanBoundaries();*/ }
  
  return rvalue;
}

//**************************************************************************
//                C A L C I U M   2 D   F I   S T E P
//**************************************************************************
//
//  (1 - Ax) Ca*  = (1 + Ay) Ca + dt H(Ca,B)
//  (1 - Ay) Ca^  = Ca* - Ay Ca + dt ( H(Ca^,B) - H(Ca,B) )
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) } + dt I
//
//  Ca = Ca(t_n), Ca^ = Ca(t_{n+1}), B = B(t_{n+1/2})
//**************************************************************************

int Ca2DFIstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int       b, rvalue = RETURN_NORMAL;
  double    t0 = Ca.Time;
  VectorObj Lin(FieldObj::Size);

  if ( (T > 0.0) && (t0 + dt > T))  { rvalue = RETURN_END; dt = T - t0; }

  double nu = Ca.getD() * dt;

  Next = Ca;
  Lin = 0.0; 
  double sum = 0.0;  
  for (b = 0; b < Buf.buf_num; b++)
    {
    Lin  -= (dt * Buf.array[b]->kplus) * *Buf.array[b];
    Next -= *Buf.array[b] * (dt * Buf.array[b]->kminus);
    sum += dt * (Buf.array[b])->kminus * (Buf.array[b])->total;
    }

  Next += sum + Lin % Ca;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);

  switch (GEOMETRY) {
  case CONICAL:   Ca.RunConR( -nu,  0.0,   nu,  0,  Next.elem); break;
  case CYLINDRICAL:   Ca.RunCylR( -nu,  0.0,   nu,  0,  Next.elem); break;
  }

  Next -= Lin % Ca;

  switch (GEOMETRY) {
  case CONICAL: Ca.RunConTheta(-nu, 0.0, -nu,  Lin.elem,  Next.elem); break;
  case CYLINDRICAL: Ca.RunCylZ(-nu, 0.0, -nu,  Lin.elem,  Next.elem); break;
  }

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; /*Ca.cleanBoundaries();*/ }
    else Ca.Time = t0;

  return rvalue;
}

//**************************************************************************
//               B U F F E R   2 D   F I   S T E P
//**************************************************************************
//
//  (1 - Ax) B*  = (1 + Ay) B + dt H(Ca,B)
//  (1 - Ay) B^  = B* - Ay B  + dt ( H(Ca,B^) - H(Ca,B) )
//
//  H(Ca,B) = - kplus * Ca * B + kminus * (B_total - B)
//
//  B = B(t_n), B^ = B(t_{n+1}), Ca = Ca(t_{n+1/2})
//**************************************************************************

int Buf2DFIstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{ 
int    rvalue = RETURN_NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T))  { dt = T - Buf.Time; rvalue = RETURN_END; }
  double nu = Buf.getD() * dt;
  
  VectorObj Lin( (-dt) * (Buf.kplus * Ca + Buf.kminus) );

  Next = Buf + dt * Buf.kminus * Buf.total + Lin % Buf;

  switch (GEOMETRY) {
  case CONICAL:  Buf.RunConR(    -nu,     0.0,   nu,  0,  Next.elem); break;
  case CYLINDRICAL:  Buf.RunCylR(    -nu,     0.0,   nu,  0,  Next.elem); break;
  }

  Next -= Lin % Buf;
  
  switch (GEOMETRY) {
  case CONICAL:  Buf.RunConTheta(-nu, 0.0, -nu,  Lin.elem,  Next.elem); break;
  case CYLINDRICAL:  Buf.RunCylZ(-nu, 0.0, -nu,  Lin.elem,  Next.elem); break;
  }

  if (T > 0.0) { Buf.Time += dt; Buf = Next; /*Buf.cleanBoundaries();*/  }
  
  return rvalue;
}

//**************************************************************************
//             C A L C I U M   L I N E A R   D - G   S T E P
//**************************************************************************
//
//  (1 - Ax/2) Ca*  = (1 + Ax/2 + Ay + Az) Ca + dt H(Ca,B)
//  (1 - Ay/2) Ca** = -Ay/2 Ca + Ca*
//  (1 - Az/2) Ca^  = -Az/2 Ca + Ca** + dt / 2 * { H(Ca^,B) - H(Ca,B) }
//
//  H(Ca,B) = Sum_B{ - kplus * (Ca * B0 + Ca0 * B) - kminus * B) }
//
//  Ca = Ca(t_n) - Ca0, Ca^ = Ca(t_{n+1}) - Ca0, B = B(t_{n+1/2}) - B0
//**************************************************************************

int CaLinADIstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int       b, rvalue = RETURN_NORMAL;
  double    t0 = Ca.Time;
  VectorObj Lin(FieldObj::Size);

  if ( (T > 0.0) && (t0 + dt > T))  {
    //printf("Ca: t0=%g dt=%g T=%g: dt reduced\n", t0,dt,T); 
    rvalue = RETURN_END; 
    dt = T - t0; 
    }
  double nu = Ca.getD() * dt;

  Next = Ca - Ca.total;
  double cB, cCa, sum = 0;

  for (b = 0; b < Buf.buf_num; b++)
    {
    cB = dt * ( Buf.array[b]->kminus + Buf.array[b]->kplus * Ca.total );
    cCa = dt * Buf.array[b]->kplus * Buf.array[b]->bgr;
    Next -= *Buf.array[b] * cB + Ca * cCa;
    sum  += ( Buf.array[b]->bgr * cB + Ca.total * cCa );
    }

  Next += sum;

  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);

  Ca.xRun(-0.5 * nu, 0.5 * nu,    nu,     nu,  0,  Next.elem);
  Ca.yRun(-0.5 * nu,     0.,  -0.5 * nu,  0.,  0,  Next.elem);

  sum = 0.0; 
  for (b = 0; b < Buf.buf_num; b++)
    sum  += (0.5 * dt * Buf.array[b]->kplus) * Buf.array[b]->bgr;

  Lin = -sum;
  Next += sum * (Ca - Ca.total);

  Ca.zRun(-0.5 * nu, 0., 0., -0.5 * nu, Lin.elem, Next.elem);

  if (T > 0.0) { Ca = Next + Ca.total; Ca.Time += 0.5 * dt; /*Ca.cleanBoundaries();*/ }
    else Ca.Time = t0;

  return rvalue;
}

//**************************************************************************
//            B U F F E R   L I N E A R   D - G   S T E P
//**************************************************************************

int BufLinADIstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{ 
VectorObj Lin(FieldObj::Size);
int    rvalue = RETURN_NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) 
    { dt = T - Buf.Time; /*printf("Buf: dt reduced to %g\n", dt);*/ rvalue = RETURN_END; }
  double nu = Buf.getD() * dt;
  
  double cB = dt * ( Buf.kminus + Buf.kplus * Ca.total );
  double cCa = dt * Buf.kplus * Buf.bgr;

  Next = Buf * (1 - cB) - Ca * cCa - ( Buf.bgr * (1 - cB) - Ca.total * cCa );

  Buf.xRun(-0.5 * nu, 0.5 * nu, nu, nu, 0, Next.elem);
  Buf.yRun(-0.5 * nu, 0., -0.5 * nu, 0., 0, Next.elem);

  Lin = -0.5 * cB;
  Next += 0.5 * cB * (Buf - Buf.bgr);
  Buf.zRun(-0.5 * nu, 0., 0., -0.5 * nu,  Lin.elem, Next.elem);

  if (T > 0.0) { Buf.Time += dt; Buf = Next + Buf.bgr; /*Buf.cleanBoundaries();*/ }

  return rvalue;
}

//**************************************************************************
//         C A L C I U M   F U L L Y   I M P L I C I T    S T E P
//**************************************************************************
//
//  (1 - Ax) Ca*  = (1 + Ay + Az) Ca + dt H(Ca,B)
//  (1 - Ay) Ca** = -Ay Ca + Ca*
//  (1 - Az) Ca^  = -Az Ca + Ca** + dt * { H(Ca^,B) - H(Ca,B) }  - no 1/2 !
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }
//
//  Ca = Ca(t_n), Ca^ = Ca(t_{n+1}), B = B(t_{n+1/2})
//**************************************************************************

int CaFIADIstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int    b, rvalue = RETURN_NORMAL;
  double t0 = Ca.Time;
  VectorObj Lin(FieldObj::Size); 
  
  if ( (T > 0.0) && (t0 + dt > T)) { dt = T - t0; rvalue = RETURN_END; }
  double nu = Ca.getD() * dt;  

  Next = Ca;     // take care of non-updating values at Dirichlet boundary
  double sum = 0.0;  
  for (b = 0; b < Buf.buf_num; b++)
    {
    Next -= *Buf.array[b] % ( dt * Buf.array[b]->kminus + (dt * Buf.array[b]->kplus) * Ca );
    sum += dt * Buf.array[b]->kminus * Buf.array[b]->total;
    }

  Next += sum;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);

  Ca.xRun(-nu, 0.0, nu, nu,  0,   Next.elem);
  Ca.yRun(-nu, 0.0, -nu,  0.,  0, Next.elem);

  Lin = 0.0; 
  for (b = 0; b < Buf.buf_num; b++)
  Lin  -= (dt * Buf.array[b]->kplus) * *Buf.array[b];  //<- no factor of 1/2 

  Next -= (Lin % Ca);
  Ca.zRun(-nu, 0., 0., -nu, Lin.elem, Next.elem);

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; /*Ca.cleanBoundaries();*/ } 
  else Ca.Time = t0;

  return rvalue;
}

//******************************************************************************************
//                B U F F E R   F U L L Y   I M P L I C I T    S T E P
//******************************************************************************************
//
//  (1 - Ax) Buf*  = (1 + Ay + Az) Buf + dt H(Buf,B)
//  (1 - Ay) Buf** = -Ay Buf + Buf*
//  (1 - Az) Buf^  = -Az Buf + Buf** + dt * { H(Buf^,Ca) - H(Buf,Ca) }  - no 1/2 !
//
//  H(Buf,B) = - kplus * Buf * Ca + kminus * (B_total - Buf) }
//
//  Buf = Buf(t_n), Buf^ = Buf(t_{n+1}), Ca = Ca(t_{n+1/2})
//**************************************************************************

int BufFIADIstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{
  int    rvalue = RETURN_NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) { dt = T - Buf.Time; rvalue = RETURN_END; }
  double nu = Buf.getD() * dt;

  Next = Buf % ((1.0 - dt * Buf.kminus) - (dt * Buf.kplus) * Ca) + dt * Buf.kminus * Buf.total;

  Buf.xRun(-nu, 0.0, nu, nu, 0,  Next.elem);
  Buf.yRun(-nu, 0.0, -nu, 0., 0, Next.elem);

  VectorObj Lin( -dt * ( Buf.kplus * Ca + Buf.kminus ) );
  Next -= (Lin % Buf);
  Buf.zRun(-nu, 0.0, 0.0, -nu,  Lin.elem, Next.elem);
  if (T > 0.0) { Buf = Next; Buf.Time += dt; /*Buf.cleanBoundaries();*/ }

  return rvalue;
}

//**************************************************************************
//              C A L C I U M   E U L E R   S T E P
//**************************************************************************
//
//  Ca(t_{n+1}) = (1 + Ax + Ay + Az) Ca(t_n) + dt H(Ca(t_n), B(t_n))
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }
//
//**************************************************************************

int CaEulerStep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int    rvalue = RETURN_NORMAL;
  double t0 = Ca.Time;

  if ( (T > 0.0) && (t0 + dt > T)) { dt = T - t0; rvalue = RETURN_END; }
  double nu = Ca.getD() * dt;

  double sum = 0.0;  

  Next = Ca;     // take care of non-updating values at Dirichlet boundary
  for (int b = 0; b < Buf.buf_num; b++)
    {
    Next -= *Buf.array[b] % ( dt * Buf.array[b]->kminus + (dt * Buf.array[b]->kplus) * Ca );
    sum += dt * Buf.array[b]->kminus * Buf.array[b]->total;
    }
  Next += sum;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);
  Ca.xRun(0.0, nu, nu, nu, 0, Next.elem);
  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; /*Ca.cleanBoundaries();*/ } 
  else Ca.Time = t0;

  return rvalue;
}

//**************************************************************************
//            B U F F E R   E U L E R   S T E P
//**************************************************************************

int BufEulerStep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{
  int    rvalue = RETURN_NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) { dt = T - Buf.Time; rvalue = RETURN_END; }
  double nu = Buf.getD() * dt;

  Next = Buf % ( 1.0 - dt * Buf.kminus - (dt * Buf.kplus) * Ca ) + dt * Buf.kminus * Buf.total;
 
  Buf.xRun(0.0, nu, nu, nu, 0, Next.elem);
  if (T > 0.0) { Buf = Next; Buf.Time += dt; /*Buf.cleanBoundaries();*/ }

  return rvalue;
}

//**************************************************************************
//              C A L C I U M   2 D   E U L E R   S T E P
//**************************************************************************
//
//  Ca(t_{n+1}) = (1 + Ax + Ay) Ca(t_n) + dt H(Ca(t_n), B(t_n))
//
//  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }
//
//**************************************************************************

int CaEu2Dstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int    rvalue = RETURN_NORMAL;
  double t0 = Ca.Time;

  if ( (T > 0.0) && (t0 + dt > T)) { dt = T - t0; rvalue = RETURN_END; }
  double nu = Ca.getD() * dt;

  double sum = 0.0;  

  Next = Ca;     // take care of non-updating values at Dirichlet boundary
  for (int b = 0; b < Buf.buf_num; b++)
    {
    Next -= *Buf.array[b] % ( dt * Buf.array[b]->kminus + (dt * Buf.array[b]->kplus) * Ca );
    sum += dt * Buf.array[b]->kminus * Buf.array[b]->total;
    }
  Next += sum;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);

  switch (GEOMETRY) {
  case CONICAL: Ca.RunConTheta(0.0, nu, nu, 0, Next.elem); break;
  case CYLINDRICAL: Ca.RunCylR(0.0, nu, nu, 0, Next.elem); break;
  }

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; /*Ca.cleanBoundaries();*/ } else Ca.Time = t0;

  return rvalue;
}

//**************************************************************************
//            B U F F E R   2 D   E U L E R   S T E P
//**************************************************************************

int BufEu2Dstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{
  int    rvalue = RETURN_NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) { dt = T - Buf.Time; rvalue = RETURN_END; }
  double nu = Buf.getD() * dt;

  Next = Buf % ( 1.0 - dt * Buf.kminus - (dt * Buf.kplus) * Ca ) + dt * Buf.kminus * Buf.total;
 
    switch (GEOMETRY) {
    case CONICAL: Buf.RunConTheta(0.0, nu, nu, 0, Next.elem); break;
    case CYLINDRICAL: Buf.RunCylR(0.0, nu, nu, 0, Next.elem); break;
    }

  if (T > 0.0) { Buf = Next; Buf.Time += dt; /*Buf.cleanBoundaries();*/ }

  return rvalue;
}

/**************************************************************************
     C A L C I U M   C R A N K - N I C H O L S O N   1 D   S T E P
***************************************************************************

  (1 - Ax/2) Ca^ = (1 + Ax/2) Ca + dt/2 ( H(Ca, B) + H(Ca^,B) ) + dt sources(t_{n+1/2})
                 = (1 + Ax/2) Ca - dt/2 Sum_B{ kplus * B } ( Ca + Ca^ )
                 + dt { Sum_B{ kminus * (B_total - B) + sources(t_{n+1/2}) }

  ( 1 - Ax/2 - Lin ) Ca^ = Ax/2 Ca + (1 + Lin ) Ca
                         + dt { Sum_B{ kminus * (B_total - B) + sources(t_{n+1/2}) }

  Lin = - dt/2 Sum_B{ kplus * B }


  H(Ca,B) = Sum_B{ - kplus * Ca * B + kminus * (B_total - B) }

  Ca = Ca(t_n), Ca^ = Ca(t_{n+1}), B = B(t_{n+1/2})
****************************************************************************/

int Ca1Dstep(FieldObj &Ca, VectorObj &Next, BufferArray &Buf, double dt, double T)
{
  int       rvalue = RETURN_NORMAL;
  double    t0 = Ca.Time;
  VectorObj Lin(FieldObj::Size);

  if ( (T > 0.0) && (t0 + dt > T))  { rvalue = RETURN_END; dt = T - t0; }

  double nu = Ca.getD() * dt;

  Next = Ca;
  double sum = 0.0;  
  Lin = 0.0; 
  for (int b = 0; b < Buf.buf_num; b++) {
    Lin  -= (0.5 * dt * Buf.array[b]->kplus) * *Buf.array[b];
    Next -= ( dt * Buf.array[b]->kminus ) * *Buf.array[b];
    sum += dt * (Buf.array[b])->kminus * (Buf.array[b])->total;
    }

  Next += sum + Ca % Lin;
  Ca.Time += 0.5 * dt;
  Ca.add_sources(Next, dt);
  Ca.RunSpheR( -0.5 * nu, 0.5 * nu, Lin.elem, Next.elem);

  if (T > 0.0) { Ca = Next; Ca.Time += 0.5 * dt; /*Ca.cleanBoundaries();*/ }
    else Ca.Time = t0;

  return rvalue;
}

/**************************************************************************
       B U F F E R    C R A N K - N I C H O L S O N   1 D   S T E P
**************************************************************************

  (1 - Ax/2) B^ = (1 + Ax/2) B + dt/2 ( H(Ca, B) + H(Ca,B^) )
                = (1 + Ax/2) B - dt/2 kplus Ca ( B + B^ ) + dt kminus (B_total - B) 

  ( 1 - Ax/2 - Lin ) B^ = Ax/2 Ca + (1 + Lin ) B + dt kminus (B_total - B)

  Lin = - dt/2 kplus Ca

  H(Ca,B) = - kplus * Ca * B + kminus * (B_total - B)

  B = B(t_n), B^ = B(t_{n+1}), Ca = Ca(t_{n+1/2})
****************************************************************************/

int Buf1Dstep(BufferObj &Buf, VectorObj &Next, FieldObj &Ca, double dt, double T)
{
  int    rvalue = RETURN_NORMAL;

  if ( (T > 0.0) && (Buf.Time + dt > T)) { dt = T - Buf.Time; rvalue = RETURN_END; }
  double nu = Buf.getD() * dt;

  VectorObj Lin( (-0.5 * dt * Buf.kplus) * Ca ); 
  Next = Buf % ( 1.0 + Lin ) + (dt * Buf.kminus) * (Buf.total - Buf);

  Buf.RunSpheR( -0.5 * nu, 0.5 * nu, Lin.elem, Next.elem);

  if (T > 0.0) { Buf = Next; Buf.Time += dt; /*Buf.cleanBoundaries();*/ }

  return rvalue;
}

//**************************************************************************

const char *getMethod( char *smethod, CaMethod &CaStep, BufMethod &BufStep) {

 switch (GEOMETRY)  { 
 case CARTESIAN3D:                // 3D cartesian geometry
     if ( equal(smethod,"1D") )
       { CaStep = &Ca1Dstep; BufStep = &Buf1Dstep;         return "radially symmetric ADI"; }
     else if ( equal(smethod,"LIN") )
       { CaStep = &CaLinADIstep; BufStep = &BufLinADIstep; return "linearized ADI"; }
     else if ( equal(smethod,"FI") )
       { CaStep = &CaFIADIstep; BufStep = &BufFIADIstep;   return "fully implicit"; }
     else if ( equal(smethod,"EU") )
       { CaStep = &CaEulerStep; BufStep = &BufEulerStep;   return "Euler"; }
     else if ( equal(smethod,"DG") || equal(smethod,"default") ) // default
       { CaStep = &CaADIstep; BufStep = &BufADIstep;       return "Douglas-Gunn ADI"; }
     break;

 case SPHERICAL:                  // ** SPHERICAL
     if ( equal(smethod,"CN") || equal(smethod,"default") )
       { CaStep = &Ca1Dstep; BufStep = &Buf1Dstep;     return "spherical 1D Crank-Nicholson implicit"; }
     break;

 case CYLINDRICAL:                // *** CYLINDRICAL
     if ( equal(smethod,"DG") || equal(smethod,"default") )
       { CaStep = &Ca2Dstep; BufStep = &Buf2Dstep;     return "cylindrical (r,z) 2D Crank-Nicholson implicit"; }
     else  if ( equal(smethod,"EU") )
       { CaStep = &CaEu2Dstep; BufStep = &BufEu2Dstep; return "cylindrical (r,z) 2D Euler"; }
     else  if ( equal(smethod,"FI") )
       { CaStep = &Ca2DFIstep; BufStep = &Buf2DFIstep; return "cylindrical (r,z) 2D fully implicit"; }
     break;
 case CONICAL:                    // *** CONICAL or 
     if ( equal(smethod,"DG") || equal(smethod,"default") )
       { CaStep = &Ca2Dstep; BufStep = &Buf2Dstep;     return "conical (r,theta) 2D Crank-Nicholson implicit"; }
     else  if ( equal(smethod,"EU") )
       { CaStep = &CaEu2Dstep; BufStep = &BufEu2Dstep; return "conical (r,theta) 2D Euler"; }
     else  if ( equal(smethod,"FI") )
       { CaStep = &Ca2DFIstep; BufStep = &Buf2DFIstep; return "conical (r,theta) 2D fully implicit"; }
     break;
 }

 throw makeMessage("Cannot recognize difference scheme \"%s\"", smethod);
}

//@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@


void getRun( TokenString &TS, int i, bool *adaptive, char *PDEmethod, double *time, double *dt, double *dtMax,
     double *accuracy, double *accuracy0, double *maxError, double *dtStretch, 
     int *n0, double *nStretch, int *nMax, double *ODEaccuracy)
{
  char firstArg[MAX_TOKEN_LENGTH], secondArg[MAX_TOKEN_LENGTH];

  if (adaptive) *adaptive = false;             // non-adaptive is the default
  long pos = TS.token_index("Run", i);

  TS.trail_pars(pos, 's', firstArg, 's', secondArg, 'E');
  pos ++;
  int pmax = TS.tokens_to_eol(pos);

  if ( pmax > 0 && !TS.isConst(firstArg) ) { 
    pos++; 
    if (adaptive) if ( equal(firstArg,"adaptive") ) *adaptive = true;
                  else if (PDEmethod) strcpy(PDEmethod, firstArg);
    else if (PDEmethod) strcpy(PDEmethod, firstArg);

    if ( pmax > 1 && !TS.isConst(secondArg) ) { 
         pos++; 
         if (PDEmethod) strcpy(PDEmethod, secondArg);
    }
  }

  if (time) *time = TS.get_double(pos);

  TS.trail_pars(pos, 'd', dt, 'd', dtMax, 'd', accuracy, 'd', accuracy0, 'd', maxError, 'd', dtStretch,
		'i', n0, 'd', nStretch, 'i', nMax, 'd', ODEaccuracy, 'E');
}


//**************************************************************************

double get_sim_time(TokenString &params)
  {
  double tt, t = 0.0;

  for (int i = 1; i<= params.token_count("Run"); i++) {
    getRun(params, i, 0, 0, &tt); 
    t += tt;
  }

  return t;
  }


//**************************************************************************


void setTortuosity(TokenString &TS, long p, double *Diff) {

  int xsize = FieldObj::Grid->xsize;  double *xcoord = FieldObj::Grid->xcoord;
  int ysize = FieldObj::Grid->ysize;  double *ycoord = FieldObj::Grid->ycoord;
  int zsize = FieldObj::Grid->zsize;  double *zcoord = FieldObj::Grid->zcoord;

  int  ix, iy, iz;
  long l = 0;
  double x, y, z;
  ExpressionObj *T = 0;

  switch (GEOMETRY) {

  case CARTESIAN3D:  // CARTESIAN
    T = new ExpressionObj(TS, p, "Bad tortuosity expression", 0, 0, &x, "x", &y, "y", &z, "z"); 
    if (VERBOSE) {
      fprintf(stdlog,"    tortuosity = ");
      T->print(stdlog); fprintf(stdlog,"\n");
    }

    for (iz = 0; iz < zsize; iz++) {
      z = zcoord[iz];
      for (iy = 0; iy < ysize; iy++) {
        y = ycoord[iy];
        for (ix = 0; ix < xsize; ix++)  {
          x = xcoord[ix];
          Diff[l++] = T->Evaluate();
	}
      }
    }
    break;  

  case SPHERICAL: // SPHERICAL COORDINATES
    T = new ExpressionObj(TS, p, "Bad tortuosity expression (function of r)", 0, 0, &x, "r"); 
    fprintf(stdlog,"\n#### Tortuosity function = ");
    T->print(stdlog); fprintf(stdlog,"\n");

    for (ix = 0; ix < xsize; ix++)  {
          x = xcoord[ix];
          Diff[l++] = T->Evaluate();
    }
    break;  

  case CONICAL: // CONICAL COORDINATES 
    T = new ExpressionObj(TS, p, "Bad tortuosity expression (function of r, theta)", 0, 0, &x, "r", &y, "theta"); 
    fprintf(stdlog,"\n#### Tortuosity function = ");
    T->print(stdlog); fprintf(stdlog,"\n");

    for (iy = 0; iy < ysize; iy++) {
        y = ycoord[iy];
        for (ix = 0; ix < xsize; ix++)  {
          x = xcoord[ix];
          Diff[l++] = T->Evaluate();
	}
    }
    break;  

  case CYLINDRICAL: // CYLINDRICAL COORDINATES 
    T = new ExpressionObj(TS, p, "Bad tortuosity expression (function of r, z)", 0, 0, &x, "r", &y, "z"); 
    fprintf(stdlog,"\n#### Tortuosity function = ");
    T->print(stdlog); fprintf(stdlog,"\n");

    for (iy = 0; iy < ysize; iy++) {
        y = ycoord[iy];
        for (ix = 0; ix < xsize; ix++)  {
          x = xcoord[ix];
          Diff[l++] = T->Evaluate();
	}
    }
    break;  
  }

 if (T) delete T;
 return;
}


//**************************************************************************
